import json

import matplotlib.pyplot as plt
from bin.signal_processing import levenshteinDistance

from bin.OCR_dataloader_iam import DataLoaderIAM, Batch
from bin.OCR_model import Model, DecoderType
from bin.OCR_preprocessor import Preprocessor


# class FilePaths:
#     """Filenames and paths to data."""
#     fn_char_list = '../OCR_model/charList.txt'
#     fn_summary = '../OCR_model/summary.json'
#     fn_corpus = '../data/corpus.txt'




def get_img_height():
    """Fixed height for NN."""
    return 32


def get_img_size(line_mode = False):
    """Height is fixed for NN, width is set according to training mode (single words or text lines)."""
    if line_mode:
        return 256, get_img_height()
    return (128, get_img_height())


def write_summary(char_error_rates, word_accuracies, FileNames):
    """Writes training summary file for NN."""
    with open(FileNames.fn_summary, 'w') as f:
        json.dump({'charErrorRates': char_error_rates, 'wordAccuracies': word_accuracies}, f)


def char_list_from_file(FileNames):
    with open(FileNames.fn_char_list) as f:
        liste = list(f.read())
        if(liste[-1] == "\n"):
            return liste[:-1]
        else:
            return liste


def train(model,
          loader,
          line_mode,
          FileNames,
          early_stopping = 25):
    """Trains NN."""
    epoch = 0  # number of training epochs since start
    summary_char_error_rates = []
    summary_word_accuracies = []
    preprocessor = Preprocessor(get_img_size(line_mode), data_augmentation=True, line_mode=line_mode)
    best_char_error_rate = float('inf')  # best validation character error rate
    no_improvement_since = 0  # number of epochs no improvement of character error rate occurred
    # stop training after this number of epochs without improvement
    while True:
        epoch += 1
        print('Epoch:', epoch)

        # train
        print('Train NN')
        loader.train_set()
        while loader.has_next():
            iter_info = loader.get_iterator_info()
            batch = loader.get_next()
            batch = preprocessor.process_batch(batch)
            loss = model.train_batch(batch)
            print("Epoch: {epoch} Batch: {iter_info[0]}/{iter_info[1]} Loss: {loss}")

        # validate
        char_error_rate, word_accuracy = validate(model, loader, line_mode)

        # write summary
        summary_char_error_rates.append(char_error_rate)
        summary_word_accuracies.append(word_accuracy)
        write_summary(summary_char_error_rates, summary_word_accuracies, FileNames)

        # if best validation accuracy so far, save model parameters
        if char_error_rate < best_char_error_rate:
            print('Character error rate improved, save model')
            best_char_error_rate = char_error_rate
            no_improvement_since = 0
            model.save()
        else:
            print("Character error rate not improved, best so far: ",char_error_rate * 100.0,"%")
            no_improvement_since += 1

        # stop training if no more improvement in the last x epochs
        if no_improvement_since >= early_stopping:
            print("No more improvement since {early_stopping} epochs. Training stopped.")
            break


def validate(model, loader, line_mode):
    """Validates NN."""
    print("Validate NN")
    loader.validation_set()
    preprocessor = Preprocessor(get_img_size(line_mode), line_mode=line_mode)
    num_char_err = 0
    num_char_total = 0
    num_word_ok = 0
    num_word_total = 0
    while loader.has_next():
        iter_info = loader.get_iterator_info()
        print("Batch: ",iter_info[0] / iter_info[1])
        batch = loader.get_next()
        batch = preprocessor.process_batch(batch)
        recognized, _ = model.infer_batch(batch)

        print("Ground truth -> Recognized")
        for i in range(len(recognized)):
            num_word_ok += 1 if batch.gt_texts[i] == recognized[i] else 0
            num_word_total += 1
            dist = levenshteinDistance(recognized[i], batch.gt_texts[i])
            num_char_err += dist
            num_char_total += len(batch.gt_texts[i])
            print('[OK]' if dist == 0 else '[ERR:%d]' % dist, '"' + batch.gt_texts[i] + '"', '->',
                  '"' + recognized[i] + '"')

    # print validation result
    char_error_rate = num_char_err / num_char_total
    word_accuracy = num_word_ok / num_word_total
    print("Character error rate: {char_error_rate * 100.0}%. Word accuracy: {word_accuracy * 100.0}%.")
    return (char_error_rate, word_accuracy)


def infer(model, fn_img):
    """Recognizes text in image provided by file path."""
    if(type(fn_img) == str):
        img = plt.imread(fn_img)
        img = (img[:,:,0]+img[:,:,1]+img[:,:,2])*255//3
        fn_img = img
        assert fn_img is not None

    preprocessor = Preprocessor(get_img_size(), dynamic_width=True, padding=16)
    img = preprocessor.process_img(fn_img)

    batch = Batch([img], None, 1)
    recognized, probability = model.infer_batch(batch, True)
    return recognized[0]


def ocr_run(FileNames, img_file='../data/word.png', mode="infer", decoder="bestpath", batch_size=100, data_dir=False, fast=False, line_mode=False, early_stopping=25, dump=False):
    """Main function.
    parser.add_argument('--mode', choices=['train', 'validate', 'infer'], default='infer')
    parser.add_argument('--decoder', choices=['bestpath', 'beamsearch', 'wordbeamsearch'], default='bestpath')
    parser.add_argument('--batch_size', help='Batch size.', type=int, default=100)
    parser.add_argument('--data_dir', help='Directory containing IAM dataset.', required=False)
    parser.add_argument('--fast', help='Load samples from LMDB.', action='store_true')
    parser.add_argument('--line_mode', help='Train to read text lines instead of single words.', action='store_true')
    parser.add_argument('--img_file', help='Image used for inference.', default='../data/word.png')
    parser.add_argument('--early_stopping', help='Early stopping epochs.', type=int, default=25)
    parser.add_argument('--dump', help='Dump output of NN to CSV file(s).', action='store_true')
    """

    # parse arguments and set CTC decoder
    decoder_mapping = {'bestpath': DecoderType.BestPath,
                       'beamsearch': DecoderType.BeamSearch,
                       'wordbeamsearch': DecoderType.WordBeamSearch}
    decoder_type = decoder_mapping[decoder]

    # train the model
    if mode == 'train':
        loader = DataLoaderIAM(data_dir, batch_size, fast=fast)

        # when in line mode, take care to have a whitespace in the char list
        char_list = loader.char_list
        if line_mode and ' ' not in char_list:
            char_list = [' '] + char_list

        # save characters and words
        with open(FileNames.fn_char_list, 'w') as f:
            f.write(''.join(char_list))

        with open(FileNames.fn_corpus, 'w') as f:
            f.write(' '.join(loader.train_words + loader.validation_words))

        model = Model(char_list, FileNames=FileNames, decoder_type=decoder_type)
        train(model, loader, line_mode=line_mode, early_stopping=early_stopping, FileNames = FileNames)

        return 0

    # evaluate it on the validation set
    elif mode == 'validate':
        loader = DataLoaderIAM(data_dir, batch_size, fast=fast)
        model = Model(char_list_from_file(FileNames), FileNames=FileNames, decoder_type=decoder_type, must_restore=True)
        validate(model, loader, line_mode)

        return 0

    # infer text on test image
    elif mode == 'infer':
        
        model = Model(char_list_from_file(FileNames), FileNames=FileNames, decoder_type=decoder_type, must_restore=True, dump=dump)
        recognized = infer(model, img_file)
        del model
        return recognized


if __name__ == '__main__':
    ocr_run()
